
/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * debug section: 6
 */

#include "squidwall.h"
#include "thread.h"
#include "request.h"

/* dummy request for logging from main.c / thread.c */
extern request_t *r;

/**
 * thread management
 *
 * - the minimum number of squidwall threads is defined by MIN_SQUIDWALL_THREADS
 * - if needed, we just create new threds
 * - we never destroy threads, so we don't have to restart them later
 */
static void thread_start_one_worker(void);

/**
 * ...
 */
static pthread_cond_t havework_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t havework_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * the squidwall threads are created with attribute detached
 */
static pthread_attr_t attr_detached;

/**
 * protected by nr_squidwall_threads_mutex
 */
volatile int nr_squidwall_threads = 0;
static pthread_mutex_t nr_squidwall_threads_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * this list maintains the idle threads
 */
static LIST_HEAD(squidwall_idle_list);
static pthread_mutex_t squidwall_idle_list_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * pointer for passing the work to some worker
 */
static struct squidwall_work *work;
static pthread_mutex_t work_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * A structure for passing work to a squidwall thread. Also for passing
 * state information between squidwall threads.
 *
 * Protected by nr_squidwall_threads_mutex.
 */
struct squidwall_work {
	struct list_head list;    /* when idle, on squidwall_idle_list */
	int nr;                   /* thread number */
	void (*fn)(request_t *r); /* callback function */
	pthread_t id;             /* if of the thread in this queue */
	request_t R[20];
};

/* EXPORT */
void thread_init(void)
{
	int i;

	/* initialize an attribute for creating detached threads */
	pthread_attr_init(&attr_detached);
	pthread_attr_setdetachstate(&attr_detached, PTHREAD_CREATE_DETACHED);

	/* start our minimum of threads */
	for (i=0; i<NUMTHREADS; i++) {
		debug(6,3, "init(), starting worker: %d!", i);
		thread_start_one_worker();
		debug(6,3, "init(), worker %d startet.", i);
	}

	return;
}

/**
 * thread_operation() - wake up a squidwall thread and put some work to it
 *
 * @param fn function to execute
 * @param arg argument to the function
 * @return
 *   0 = everything was okay
 *  -1 = some error detected
 */
int thread_operation(void (*fn)(request_t *r))
{
	struct squidwall_work *w;
	struct timespec t;

	/* no function given */
	if (fn == NULL) return -1;

	pthread_mutex_lock(&squidwall_idle_list_mutex);

	if (list_empty(&squidwall_idle_list)) {
		pthread_mutex_unlock(&squidwall_idle_list_mutex);
		debug(6,3," operation(): Oops! squidwall_idle_list is empty (STARTING NEW)!");
		thread_start_one_worker();
		t.tv_sec=0; t.tv_nsec=50;
		nanosleep(&t,&t); /* let the thread start */
		return thread_operation(fn);
	}

	w = list_entry(squidwall_idle_list.next, struct squidwall_work, list);
	list_del_init(&w->list); /* remove from idle list */
	debug(6,3," operation(): squidwall_idle_list, sth. found (nr=%d)!", w->nr);
	w->fn = fn;
	pthread_mutex_unlock(&squidwall_idle_list_mutex);

	debug(6,3," operation(): pthread_cond_signal(pre) (nr=%d)!", w->nr);

	pthread_mutex_lock(&work_mutex);
	work=w;

	/* give the wirk to some worker */
	pthread_cond_signal(&havework_cond);
	debug(6,3," operation(): pthread_cond_signal(post) (nr=%d)!", w->nr);

	return 0;
}


/* INTERN */
static void *thread_worker(void *r)
{
	struct squidwall_work *w, W;
	int rv=1;

	w = &W;

	INIT_LIST_HEAD(&w->list);
	w->nr=nr_squidwall_threads;
	w->id=pthread_self();
	w->fn = NULL;

	debug(6,3,"  worker() STARTED: (nr=%d)", w->nr);

	for (;;) {
		debug(6,3,"  worker(), zur idle list (nr=%d)", w->nr);
		pthread_mutex_lock(&squidwall_idle_list_mutex);
		list_move(&w->list, &squidwall_idle_list); /* add to idle list */
		pthread_mutex_unlock(&squidwall_idle_list_mutex);

		/* wait for work */
		pthread_mutex_lock(&havework_mutex);
		rv = pthread_cond_wait(&havework_cond, &havework_mutex);

		/* we got now some work */
		w = work;
		pthread_mutex_unlock(&work_mutex);
		
		debug_r(r,6,3,"  worker(), starting some work (nr=%d)", w->nr);

		if (rv) {
			debug(6,3,"  worker(), Oops (nr=%d) errno=%d", w->nr, errno);
			die_pthread();
		}

		if (w->fn == NULL) {
			debug(6,3,"  worker(), NULL work function (nr=%d)", w->nr);
			continue;
		}

		/* execute some function */
		debug(6,3,"  worker(), handle request with thread nr=%d", w->nr);
		(*w->fn)(r);
		debug(6,3,"  worker(), thread nr=%d has finished", w->nr);

		pthread_mutex_unlock(&havework_mutex);
	}

	/* never reached */
	return 0;
}

static void thread_start_one_worker(void)
{
	request_t r;
	pthread_t p;
	int rv;

	/* init new logging buffer */
	log_init(&r);

	pthread_mutex_lock(&nr_squidwall_threads_mutex);
	nr_squidwall_threads++;
	pthread_mutex_unlock(&nr_squidwall_threads_mutex);

	rv = pthread_create(&p, &attr_detached, thread_worker, (void*)&r);
	if (rv) die_pthread();

	return;
}
